@@ -224,7 +224,7 @@ static void php_phongo_log(mongoc_log_level_t log_level, const char *log_domain,
 /* }}} */
 
 /* {{{ Init objects */
-static void phongo_cursor_init(zval *return_value, mongoc_cursor_t *cursor, mongoc_client_t *client TSRMLS_DC) /* {{{ */
+static void phongo_cursor_init(zval *return_value, zval *manager, mongoc_cursor_t *cursor TSRMLS_DC) /* {{{ */
 {
 	php_phongo_cursor_t *intern;
 
@@ -233,18 +233,32 @@ static void phongo_cursor_init(zval *return_value, mongoc_cursor_t *cursor, mong
 	intern = Z_CURSOR_OBJ_P(return_value);
 	intern->cursor = cursor;
 	intern->server_id = mongoc_cursor_get_hint(cursor);
-	intern->client = client;
+	intern->client = Z_MANAGER_OBJ_P(manager)->client;
+
+#if PHP_VERSION_ID >= 70000
+	ZVAL_COPY(&intern->manager, manager);
+#else
+	Z_ADDREF_P(manager);
+	intern->manager = manager;
+#endif
 } /* }}} */
 
-void phongo_server_init(zval *return_value, mongoc_client_t *client, int server_id TSRMLS_DC) /* {{{ */
+void phongo_server_init(zval *return_value, zval *manager, int server_id TSRMLS_DC) /* {{{ */
 {
 	php_phongo_server_t *server;
 
 	object_init_ex(return_value, php_phongo_server_ce);
 
 	server = Z_SERVER_OBJ_P(return_value);
-	server->client = client;
 	server->server_id = server_id;
+	server->client = Z_MANAGER_OBJ_P(manager)->client;
+
+#if PHP_VERSION_ID >= 70000
+	ZVAL_COPY(&server->manager, manager);
+#else
+	Z_ADDREF_P(manager);
+	server->manager = manager;
+#endif
 }
 /* }}} */
 
@@ -467,34 +481,25 @@ zend_bool phongo_writeerror_init(zval *return_value, bson_t *bson TSRMLS_DC) /*
 	return true;
 } /* }}} */
 
-php_phongo_writeresult_t *phongo_writeresult_init(zval *return_value, mongoc_write_result_t *write_result, mongoc_client_t *client, int server_id TSRMLS_DC) /* {{{ */
+php_phongo_writeresult_t *phongo_writeresult_init(zval *return_value, bson_t *reply, zval *manager, int server_id TSRMLS_DC) /* {{{ */
 {
 	php_phongo_writeresult_t *writeresult;
 
 	object_init_ex(return_value, php_phongo_writeresult_ce);
 
 	writeresult = Z_WRITERESULT_OBJ_P(return_value);
-	writeresult->client    = client;
+	writeresult->reply     = bson_copy(reply);
 	writeresult->server_id = server_id;
+	writeresult->client = Z_MANAGER_OBJ_P(manager)->client;
 
-	/* Copy write_results or else it'll get destroyed with the bulk destruction */
-#define SCP(field) writeresult->write_result.field = write_result->field
-	SCP(omit_nModified);
-	SCP(nInserted);
-	SCP(nMatched);
-	SCP(nModified);
-	SCP(nRemoved);
-	SCP(nUpserted);
-
-	bson_copy_to(&write_result->upserted,           &writeresult->write_result.upserted);
-	SCP(n_writeConcernErrors);
-	bson_copy_to(&write_result->writeConcernErrors, &writeresult->write_result.writeConcernErrors);
-	bson_copy_to(&write_result->writeErrors,        &writeresult->write_result.writeErrors);
-	SCP(upsert_append_count);
-#undef SCP
+#if PHP_VERSION_ID >= 70000
+	ZVAL_COPY(&writeresult->manager, manager);
+#else
+	Z_ADDREF_P(manager);
+	writeresult->manager = manager;
+#endif
 
 	return writeresult;
-
 } /* }}} */
 /* }}} */
 
@@ -522,64 +527,18 @@ mongoc_bulk_operation_t *phongo_bulkwrite_init(zend_bool ordered) { /* {{{ */
 	return mongoc_bulk_operation_new(ordered);
 } /* }}} */
 
-static void phongo_bulk_write_error_add_message(char **tmp_msg, bson_t *errors)
-{
-	bson_iter_t iter;
-
-	bson_iter_init(&iter, errors);
-
-	while (bson_iter_next(&iter)) {
-		bson_t cbson;
-		uint32_t len;
-		const uint8_t *data;
-		bson_iter_t inner_iter;
-
-		if (!BSON_ITER_HOLDS_DOCUMENT(&iter)) {
-			continue;
-		}
-
-		bson_iter_document(&iter, &len, &data);
-
-		if (!bson_init_static(&cbson, data, len)) {
-			continue;
-		}
-
-		if (bson_iter_init_find(&inner_iter, &cbson, "errmsg") && BSON_ITER_HOLDS_UTF8(&inner_iter)) {
-			const char *tmp_errmsg = bson_iter_utf8(&inner_iter, NULL);
-			size_t tmp_errmsg_len = strlen(tmp_errmsg);
-
-			*tmp_msg = erealloc(*tmp_msg, strlen(*tmp_msg) + tmp_errmsg_len + 5);
-			strncpy(*tmp_msg + strlen(*tmp_msg), " :: ", 5);
-			strncpy(*tmp_msg + strlen(*tmp_msg), tmp_errmsg, tmp_errmsg_len + 1);
-		}
-	}
-}
-
-static char* phongo_assemble_bulk_write_error(mongoc_write_result_t *write_result)
-{
-	char *tmp_msg = emalloc(sizeof("BulkWrite error"));
-
-	strncpy(tmp_msg, "BulkWrite error", sizeof("BulkWrite error"));
-
-	if (!bson_empty0(&write_result->writeErrors)) {
-		phongo_bulk_write_error_add_message(&tmp_msg, &write_result->writeErrors);
-	}
-
-	if (!bson_empty0(&write_result->writeConcernErrors)) {
-		phongo_bulk_write_error_add_message(&tmp_msg, &write_result->writeConcernErrors);
-	}
-
-	return tmp_msg;
-}
-
-bool phongo_execute_write(mongoc_client_t *client, const char *namespace, mongoc_bulk_operation_t *bulk, const mongoc_write_concern_t *write_concern, int server_id, zval *return_value, int return_value_used TSRMLS_DC) /* {{{ */
+bool phongo_execute_write(zval *manager, const char *namespace, mongoc_bulk_operation_t *bulk, const mongoc_write_concern_t *write_concern, int server_id, zval *return_value, int return_value_used TSRMLS_DC) /* {{{ */
 {
+	mongoc_client_t *client;
 	bson_error_t error;
 	char *dbname;
 	char *collname;
 	int success;
+	bson_t reply = BSON_INITIALIZER;
 	php_phongo_writeresult_t *writeresult;
 
+	client = Z_MANAGER_OBJ_P(manager)->client;
+
 	if (!phongo_split_namespace(namespace, &dbname, &collname)) {
 		phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT TSRMLS_CC, "%s: %s", "Invalid namespace provided", namespace);
 		return false;
@@ -604,48 +563,49 @@ bool phongo_execute_write(mongoc_client_t *client, const char *namespace, mongoc
 		mongoc_bulk_operation_set_hint(bulk, server_id);
 	}
 
-	success = mongoc_bulk_operation_execute(bulk, NULL, &error);
+	success = mongoc_bulk_operation_execute(bulk, &reply, &error);
 
 	/* Write succeeded and the user doesn't care for the results */
 	if (success && !return_value_used) {
+		bson_destroy(&reply);
 		return true;
 	}
 
 	/* Check for connection related exceptions */
 	if (EG(exception)) {
+		bson_destroy(&reply);
 		return false;
 	}
 
-	writeresult = phongo_writeresult_init(return_value, &bulk->result, client, bulk->hint TSRMLS_CC);
+	writeresult = phongo_writeresult_init(return_value, &reply, manager, bulk->hint TSRMLS_CC);
 	writeresult->write_concern = mongoc_write_concern_copy(write_concern);
 
 	/* The Write failed */
 	if (!success) {
-		/* The Command itself failed */
-		if (bson_empty0(&writeresult->write_result.writeErrors) && bson_empty0(&writeresult->write_result.writeConcernErrors)) {
-			/* FIXME: Maybe we can look at write_result.error and not pass error at all? */
-			phongo_throw_exception_from_bson_error_t(&error TSRMLS_CC);
-		} else {
-			char *bulk_error_msg;
-
-			bulk_error_msg = phongo_assemble_bulk_write_error(&writeresult->write_result);
-			phongo_throw_exception(PHONGO_ERROR_WRITE_FAILED TSRMLS_CC, "%s", bulk_error_msg);
-			efree(bulk_error_msg);
+		if (error.domain == MONGOC_ERROR_COMMAND || error.domain == MONGOC_ERROR_WRITE_CONCERN) {
+			phongo_throw_exception(PHONGO_ERROR_WRITE_FAILED TSRMLS_CC, "%s", error.message);
 			phongo_add_exception_prop(ZEND_STRL("writeResult"), return_value TSRMLS_CC);
+		} else {
+			phongo_throw_exception_from_bson_error_t(&error TSRMLS_CC);
 		}
-		return false;
 	}
-	return true;
+
+	bson_destroy(&reply);
+
+	return success;
 } /* }}} */
 
-int phongo_execute_query(mongoc_client_t *client, const char *namespace, const php_phongo_query_t *query, const mongoc_read_prefs_t *read_preference, int server_id, zval *return_value, int return_value_used TSRMLS_DC) /* {{{ */
+int phongo_execute_query(zval *manager, const char *namespace, const php_phongo_query_t *query, const mongoc_read_prefs_t *read_preference, int server_id, zval *return_value, int return_value_used TSRMLS_DC) /* {{{ */
 {
+	mongoc_client_t *client;
 	const bson_t *doc = NULL;
 	mongoc_cursor_t *cursor;
 	char *dbname;
 	char *collname;
 	mongoc_collection_t *collection;
 
+	client = Z_MANAGER_OBJ_P(manager)->client;
+
 	if (!phongo_split_namespace(namespace, &dbname, &collname)) {
 		phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT TSRMLS_CC, "%s: %s", "Invalid namespace provided", namespace);
 		return false;
@@ -693,17 +653,19 @@ int phongo_execute_query(mongoc_client_t *client, const char *namespace, const p
 		return true;
 	}
 
-	phongo_cursor_init(return_value, cursor, client TSRMLS_CC);
+	phongo_cursor_init(return_value, manager, cursor TSRMLS_CC);
 	return true;
 } /* }}} */
 
-int phongo_execute_command(mongoc_client_t *client, const char *db, const bson_t *command, const mongoc_read_prefs_t *read_preference, int server_id, zval *return_value, int return_value_used TSRMLS_DC) /* {{{ */
+int phongo_execute_command(zval *manager, const char *db, const bson_t *command, const mongoc_read_prefs_t *read_preference, int server_id, zval *return_value, int return_value_used TSRMLS_DC) /* {{{ */
 {
+	mongoc_client_t *client;
 	mongoc_cursor_t *cursor;
 	const bson_t *doc;
 	bson_iter_t iter;
 	bson_iter_t child;
 
+	client = Z_MANAGER_OBJ_P(manager)->client;
 
 	cursor = mongoc_client_command(client, db, MONGOC_QUERY_NONE, 0, 1, 0, command, NULL, read_preference);
 	if (server_id > 0) {
@@ -769,7 +731,7 @@ int phongo_execute_command(mongoc_client_t *client, const char *db, const bson_t
 		mongoc_cursor_next(cursor, &doc);
 	}
 
-	phongo_cursor_init(return_value, cursor, client TSRMLS_CC);
+	phongo_cursor_init(return_value, manager, cursor TSRMLS_CC);
 	return true;
 } /* }}} */
 
@@ -1050,13 +1012,13 @@ ssize_t phongo_stream_poll (mongoc_stream_poll_t *streams, size_t nstreams, int3
 } /* }}} */
 
 #if PHP_VERSION_ID < 50600
-int php_mongo_verify_hostname(const char *hostname, X509 *cert TSRMLS_DC)
+static int php_phongo_verify_hostname(const char *hostname, X509 *cert TSRMLS_DC)
 {
-	if (php_mongo_matches_san_list(cert, hostname) == SUCCESS) {
+	if (php_mongodb_matches_san_list(cert, hostname) == SUCCESS) {
 		return SUCCESS;
 	}
 
-	if (php_mongo_matches_common_name(cert, hostname TSRMLS_CC) == SUCCESS) {
+	if (php_mongodb_matches_common_name(cert, hostname TSRMLS_CC) == SUCCESS) {
 		return SUCCESS;
 	}
 
@@ -1079,10 +1041,15 @@ int php_phongo_peer_verify(php_stream *stream, X509 *cert, const char *hostname,
 			peer = hostname;
 		}
 
-		if (php_mongo_verify_hostname(peer, cert TSRMLS_CC) == FAILURE) {
+#ifdef HAVE_OPENSSL_EXT
+		if (php_phongo_verify_hostname(peer, cert TSRMLS_CC) == FAILURE) {
 			bson_set_error(error, MONGOC_ERROR_STREAM, MONGOC_ERROR_STREAM_CONNECT, "Remote certificate SubjectAltName or CN does not match '%s'", hostname);
 			return false;
 		}
+#else
+		bson_set_error(error, MONGOC_ERROR_STREAM, MONGOC_ERROR_STREAM_CONNECT, "Cannot verify remote certificate SubjectAltName or CN. Please ensure that extension is compiled against PHP with OpenSSL or disable the \"verify_peer_name\" SSL context option.");
+		return false;
+#endif
 	}
 
 	return true;
@@ -1134,9 +1101,10 @@ bool php_phongo_ssl_verify(php_stream *stream, const char *hostname, bson_error_
 #else
 	if (php_stream_context_get_option(PHP_STREAM_CONTEXT(stream), "ssl", "verify_expiry", &verify_expiry) == SUCCESS && zend_is_true(*verify_expiry)) {
 #endif
+#ifdef HAVE_OPENSSL_EXT
 		time_t current     = time(NULL);
-		time_t valid_from  = php_mongo_asn1_time_to_time_t(X509_get_notBefore(cert) TSRMLS_CC);
-		time_t valid_until = php_mongo_asn1_time_to_time_t(X509_get_notAfter(cert) TSRMLS_CC);
+		time_t valid_from  = php_mongodb_asn1_time_to_time_t(X509_get_notBefore(cert) TSRMLS_CC);
+		time_t valid_until = php_mongodb_asn1_time_to_time_t(X509_get_notAfter(cert) TSRMLS_CC);
 
 		if (valid_from > current) {
 			bson_set_error(error, MONGOC_ERROR_STREAM, MONGOC_ERROR_STREAM_CONNECT, "Certificate is not valid yet on %s", hostname);
@@ -1146,6 +1114,10 @@ bool php_phongo_ssl_verify(php_stream *stream, const char *hostname, bson_error_
 			bson_set_error(error, MONGOC_ERROR_STREAM, MONGOC_ERROR_STREAM_CONNECT, "Certificate has expired on %s", hostname);
 			return false;
 		}
+#else
+		bson_set_error(error, MONGOC_ERROR_STREAM, MONGOC_ERROR_STREAM_CONNECT, "Cannot verify certificate expiration. Please ensure that extension is compiled against PHP with OpenSSL or disable the \"verify_expiry\" SSL context option.");
+		return false;
+#endif
 	}
 
 	return true;
@@ -1190,7 +1162,7 @@ mongoc_stream_t* phongo_stream_initiator(const mongoc_uri_t *uri, const mongoc_h
 
 	options = mongoc_uri_get_options(uri);
 
-	if (bson_iter_init_find (&iter, options, "connecttimeoutms") && BSON_ITER_HOLDS_INT32 (&iter)) {
+	if (bson_iter_init_find_case (&iter, options, "connecttimeoutms") && BSON_ITER_HOLDS_INT32 (&iter)) {
 		int32_t connecttimeoutms = MONGOC_DEFAULT_CONNECTTIMEOUTMS;
 
 		if (!(connecttimeoutms = bson_iter_int32(&iter))) {
@@ -1201,6 +1173,7 @@ mongoc_stream_t* phongo_stream_initiator(const mongoc_uri_t *uri, const mongoc_h
 		timeout.tv_usec = (connecttimeoutms % 1000) * 1000;
 
 		timeoutp = &timeout;
+		MONGOC_DEBUG("Applying connectTimeoutMS: %d", connecttimeoutms);
 	}
 
 	spprintf(&uniqid, 0, "%s:%d[%s]", host->host, host->port, mongoc_uri_get_string(uri));
@@ -1930,12 +1903,14 @@ static bool php_phongo_apply_wc_options_to_client(mongoc_client_t *client, bson_
 	return true;
 } /* }}} */
 
-static mongoc_client_t *php_phongo_make_mongo_client(const mongoc_uri_t *uri, zval *driverOptions TSRMLS_DC) /* {{{ */
+static mongoc_client_t *php_phongo_make_mongo_client(php_phongo_manager_t *manager, const mongoc_uri_t *uri, zval *driverOptions TSRMLS_DC) /* {{{ */
 {
 #if PHP_VERSION_ID >= 70000
-	zval                      *tmp;
+	zval                      *zdebug = NULL;
+	zval                      *zcontext = NULL;
 #else
-	zval                     **tmp;
+	zval                     **zdebug = NULL;
+	zval                     **zcontext = NULL;
 #endif
 	php_stream_context        *ctx = NULL;
 	const char                *mech, *mongoc_version, *bson_version;
@@ -1944,30 +1919,31 @@ static mongoc_client_t *php_phongo_make_mongo_client(const mongoc_uri_t *uri, zv
 	ENTRY;
 
 #if PHP_VERSION_ID >= 70000
-	if (driverOptions && (tmp = zend_hash_str_find(Z_ARRVAL_P(driverOptions), "debug", sizeof("debug")-1)) != NULL) {
+	if (driverOptions && (zdebug = zend_hash_str_find(Z_ARRVAL_P(driverOptions), "debug", sizeof("debug")-1)) != NULL) {
 		zend_string *key = zend_string_init(PHONGO_DEBUG_INI, sizeof(PHONGO_DEBUG_INI)-1, 0);
-		zend_string *value_str = zval_get_string(tmp);
+		zend_string *value_str = zval_get_string(zdebug);
 		zend_alter_ini_entry_ex(key, value_str, PHP_INI_USER, PHP_INI_STAGE_RUNTIME, 0);
 		zend_string_release(key);
 		zend_string_release(value_str);
 	}
 #else
-	if (driverOptions && zend_hash_find(Z_ARRVAL_P(driverOptions), "debug", strlen("debug") + 1, (void**)&tmp) == SUCCESS) {
-		convert_to_string(*tmp);
+	if (driverOptions && zend_hash_find(Z_ARRVAL_P(driverOptions), "debug", strlen("debug") + 1, (void**)&zdebug) == SUCCESS) {
+		convert_to_string(*zdebug);
 
-		zend_alter_ini_entry_ex((char *)PHONGO_DEBUG_INI, sizeof(PHONGO_DEBUG_INI), Z_STRVAL_PP(tmp), Z_STRLEN_PP(tmp), PHP_INI_USER, PHP_INI_STAGE_RUNTIME, 0 TSRMLS_CC);
+		zend_alter_ini_entry_ex((char *)PHONGO_DEBUG_INI, sizeof(PHONGO_DEBUG_INI), Z_STRVAL_PP(zdebug), Z_STRLEN_PP(zdebug), PHP_INI_USER, PHP_INI_STAGE_RUNTIME, 0 TSRMLS_CC);
 	}
 #endif
 
 #if PHP_VERSION_ID >= 70000
-	if (driverOptions && (tmp = zend_hash_str_find(Z_ARRVAL_P(driverOptions), "context", sizeof("context")-1)) != NULL) {
-		ctx = php_stream_context_from_zval(tmp, 0);
+	if (driverOptions && (zcontext = zend_hash_str_find(Z_ARRVAL_P(driverOptions), "context", sizeof("context")-1)) != NULL) {
+		ctx = php_stream_context_from_zval(zcontext, 0);
 #else
-	if (driverOptions && zend_hash_find(Z_ARRVAL_P(driverOptions), "context", strlen("context") + 1, (void**)&tmp) == SUCCESS) {
-		ctx = php_stream_context_from_zval(*tmp, 0);
+	if (driverOptions && zend_hash_find(Z_ARRVAL_P(driverOptions), "context", strlen("context") + 1, (void**)&zcontext) == SUCCESS) {
+		ctx = php_stream_context_from_zval(*zcontext, 0);
 #endif
 	} else {
-		GET_DEFAULT_CONTEXT();
+		zval *tmp = NULL; /* PHP 5.x requires an lvalue */
+		ctx = php_stream_context_from_zval(tmp, 0);
 	}
 
 	if (mongoc_uri_get_ssl(uri)) {
@@ -2014,21 +1990,23 @@ static mongoc_client_t *php_phongo_make_mongo_client(const mongoc_uri_t *uri, zv
 
 #if PHP_VERSION_ID >= 70000
 		if ((pem = php_stream_context_get_option(ctx, "ssl", "local_cert")) != NULL) {
+			zend_string *s = zval_get_string(pem);
 #else
 		if (SUCCESS == php_stream_context_get_option(ctx, "ssl", "local_cert", &pem)) {
+			convert_to_string_ex(pem);
 #endif
-			char filename[MAXPATHLEN];
+			/* mongoc_client_set_ssl_opts() copies mongoc_ssl_opt_t shallowly;
+			 * its strings must be kept valid for the life of mongoc_client_t */
+			manager->pem_file = ecalloc(1, MAXPATHLEN);
 
 #if PHP_VERSION_ID >= 70000
-			zend_string *s = zval_get_string(pem);
-			if (VCWD_REALPATH(ZSTR_VAL(s), filename)) {
+			if (VCWD_REALPATH(ZSTR_VAL(s), manager->pem_file)) {
 #else
-			convert_to_string_ex(pem);
-			if (VCWD_REALPATH(Z_STRVAL_PP(pem), filename)) {
+			if (VCWD_REALPATH(Z_STRVAL_PP(pem), manager->pem_file)) {
 #endif
-				mongoc_ssl_opt_t  ssl_options;
+				mongoc_ssl_opt_t ssl_options = {0};
 
-				ssl_options.pem_file = filename;
+				ssl_options.pem_file = manager->pem_file;
 				mongoc_client_set_ssl_opts(client, &ssl_options);
 			}
 #if PHP_VERSION_ID >= 70000
@@ -2051,7 +2029,7 @@ bool phongo_manager_init(php_phongo_manager_t *manager, const char *uri_string,
 		return false;
 	}
 
-	manager->client = php_phongo_make_mongo_client(uri, driverOptions TSRMLS_CC);
+	manager->client = php_phongo_make_mongo_client(manager, uri, driverOptions TSRMLS_CC);
 	mongoc_uri_destroy(uri);
 
 	if (!manager->client) {
@@ -2066,6 +2044,17 @@ bool phongo_manager_init(php_phongo_manager_t *manager, const char *uri_string,
 		return false;
 	}
 
+	/* Keep a reference to driverOptions, since it may be referenced later for
+	 * lazy stream initialization. */
+	if (driverOptions) {
+#if PHP_VERSION_ID >= 70000
+		ZVAL_COPY(&manager->driverOptions, driverOptions);
+#else
+		Z_ADDREF_P(driverOptions);
+		manager->driverOptions = driverOptions;
+#endif
+	}
+
 	return true;
 } /* }}} */
 
