@@ -42,6 +42,10 @@
 #include "php_phongo.h"
 #include "php_bson.h"
 
+#define RETURN_LONG_FROM_BSON_INT32(iter, bson, key) \
+	if (bson_iter_init_find((iter), (bson), (key)) && BSON_ITER_HOLDS_INT32((iter))) { \
+		RETURN_LONG(bson_iter_int32((iter))); \
+	}
 
 PHONGO_API zend_class_entry *php_phongo_writeresult_ce;
 
@@ -51,6 +55,7 @@ zend_object_handlers php_phongo_handler_writeresult;
    Returns the number of documents that were inserted */
 PHP_METHOD(WriteResult, getInsertedCount)
 {
+	bson_iter_t iter;
 	php_phongo_writeresult_t *intern;
 	SUPPRESS_UNUSED_WARNING(return_value_ptr) SUPPRESS_UNUSED_WARNING(return_value_used)
 
@@ -61,14 +66,14 @@ PHP_METHOD(WriteResult, getInsertedCount)
 		return;
 	}
 
-
-	RETURN_LONG(intern->write_result.nInserted);
+	RETURN_LONG_FROM_BSON_INT32(&iter, intern->reply, "nInserted");
 }
 /* }}} */
 /* {{{ proto integer WriteResult::getMatchedCount()
    Returns the number of documents that matched the update criteria */
 PHP_METHOD(WriteResult, getMatchedCount)
 {
+	bson_iter_t iter;
 	php_phongo_writeresult_t *intern;
 	SUPPRESS_UNUSED_WARNING(return_value_ptr) SUPPRESS_UNUSED_WARNING(return_value_used)
 
@@ -79,14 +84,14 @@ PHP_METHOD(WriteResult, getMatchedCount)
 		return;
 	}
 
-
-	RETURN_LONG(intern->write_result.nMatched);
+	RETURN_LONG_FROM_BSON_INT32(&iter, intern->reply, "nMatched");
 }
 /* }}} */
 /* {{{ proto integer|null WriteResult::getModifiedCount()
    Returns the number of documents that were actually modified by an update */
 PHP_METHOD(WriteResult, getModifiedCount)
 {
+	bson_iter_t iter;
 	php_phongo_writeresult_t *intern;
 	SUPPRESS_UNUSED_WARNING(return_value_ptr) SUPPRESS_UNUSED_WARNING(return_value_used)
 
@@ -97,17 +102,14 @@ PHP_METHOD(WriteResult, getModifiedCount)
 		return;
 	}
 
-	if (intern->write_result.omit_nModified) {
-		RETURN_NULL();
-	}
-
-	RETURN_LONG(intern->write_result.nModified);
+	RETURN_LONG_FROM_BSON_INT32(&iter, intern->reply, "nModified");
 }
 /* }}} */
 /* {{{ proto integer WriteResult::getDeletedCount()
    Returns the number of documents that were deleted */
 PHP_METHOD(WriteResult, getDeletedCount)
 {
+	bson_iter_t iter;
 	php_phongo_writeresult_t *intern;
 	SUPPRESS_UNUSED_WARNING(return_value_ptr) SUPPRESS_UNUSED_WARNING(return_value_used)
 
@@ -118,14 +120,14 @@ PHP_METHOD(WriteResult, getDeletedCount)
 		return;
 	}
 
-
-	RETURN_LONG(intern->write_result.nRemoved);
+	RETURN_LONG_FROM_BSON_INT32(&iter, intern->reply, "nRemoved");
 }
 /* }}} */
 /* {{{ proto integer WriteResult::getUpsertedCount()
    Returns the number of documents that were upserted */
 PHP_METHOD(WriteResult, getUpsertedCount)
 {
+	bson_iter_t iter;
 	php_phongo_writeresult_t *intern;
 	SUPPRESS_UNUSED_WARNING(return_value_ptr) SUPPRESS_UNUSED_WARNING(return_value_used)
 
@@ -136,8 +138,7 @@ PHP_METHOD(WriteResult, getUpsertedCount)
 		return;
 	}
 
-
-	RETURN_LONG(intern->write_result.nUpserted);
+	RETURN_LONG_FROM_BSON_INT32(&iter, intern->reply, "nUpserted");
 }
 /* }}} */
 /* {{{ proto MongoDB\Driver\Server WriteResult::getServer()
@@ -154,14 +155,18 @@ PHP_METHOD(WriteResult, getServer)
 		return;
 	}
 
-
-	phongo_server_init(return_value, intern->client, intern->server_id TSRMLS_CC);
+#if PHP_VERSION_ID >= 70000
+	phongo_server_init(return_value, &intern->manager, intern->server_id TSRMLS_CC);
+#else
+	phongo_server_init(return_value, intern->manager, intern->server_id TSRMLS_CC);
+#endif
 }
 /* }}} */
 /* {{{ proto array WriteResult::getUpsertedIds()
    Returns the identifiers generated by the server for upsert operations. */
 PHP_METHOD(WriteResult, getUpsertedIds)
 {
+	bson_iter_t iter, child;
 	php_phongo_writeresult_t *intern;
 	SUPPRESS_UNUSED_WARNING(return_value_ptr) SUPPRESS_UNUSED_WARNING(return_value_used)
 
@@ -175,16 +180,12 @@ PHP_METHOD(WriteResult, getUpsertedIds)
 
 	array_init(return_value);
 
-	if (!bson_empty0(&intern->write_result.upserted)) {
-		bson_iter_t iter;
-
-		bson_iter_init(&iter, &intern->write_result.upserted);
-
-		while (bson_iter_next(&iter)) {
+	if (bson_iter_init_find(&iter, intern->reply, "upserted") && BSON_ITER_HOLDS_ARRAY(&iter) && bson_iter_recurse(&iter, &child)) {
+		while (bson_iter_next(&child)) {
 			int32_t index;
 			bson_iter_t outer;
 
-			if (!BSON_ITER_HOLDS_DOCUMENT(&iter) || !bson_iter_recurse(&iter, &outer)) {
+			if (!BSON_ITER_HOLDS_DOCUMENT(&child) || !bson_iter_recurse(&child, &outer)) {
 				continue;
 			}
 			if (!bson_iter_find(&outer, "index") || !BSON_ITER_HOLDS_INT32(&outer)) {
@@ -227,6 +228,7 @@ PHP_METHOD(WriteResult, getUpsertedIds)
    Return any write concern error that occurred */
 PHP_METHOD(WriteResult, getWriteConcernError)
 {
+	bson_iter_t iter, child;
 	php_phongo_writeresult_t *intern;
 	SUPPRESS_UNUSED_WARNING(return_value_ptr) SUPPRESS_UNUSED_WARNING(return_value_used)
 
@@ -237,22 +239,17 @@ PHP_METHOD(WriteResult, getWriteConcernError)
 		return;
 	}
 
-
-	if (!bson_empty0(&intern->write_result.writeConcernErrors)) {
-		bson_iter_t iter;
-
-		bson_iter_init(&iter, &intern->write_result.writeConcernErrors);
-
-		while (bson_iter_next(&iter)) {
+	if (bson_iter_init_find(&iter, intern->reply, "writeConcernErrors") && BSON_ITER_HOLDS_ARRAY(&iter) && bson_iter_recurse(&iter, &child)) {
+		while (bson_iter_next(&child)) {
 			bson_t cbson;
 			uint32_t len;
 			const uint8_t *data;
 
-			if (!BSON_ITER_HOLDS_DOCUMENT(&iter)) {
+			if (!BSON_ITER_HOLDS_DOCUMENT(&child)) {
 				continue;
 			}
 
-			bson_iter_document(&iter, &len, &data);
+			bson_iter_document(&child, &len, &data);
 
 			if (!bson_init_static(&cbson, data, len)) {
 				continue;
@@ -277,6 +274,7 @@ PHP_METHOD(WriteResult, getWriteConcernError)
    Returns any write errors that occurred */
 PHP_METHOD(WriteResult, getWriteErrors)
 {
+	bson_iter_t iter, child;
 	php_phongo_writeresult_t *intern;
 	SUPPRESS_UNUSED_WARNING(return_value_ptr) SUPPRESS_UNUSED_WARNING(return_value_used)
 
@@ -290,12 +288,8 @@ PHP_METHOD(WriteResult, getWriteErrors)
 
 	array_init(return_value);
 
-	if (!bson_empty0(&intern->write_result.writeErrors)) {
-		bson_iter_t iter;
-
-		bson_iter_init(&iter, &intern->write_result.writeErrors);
-
-		while (bson_iter_next(&iter)) {
+	if (bson_iter_init_find(&iter, intern->reply, "writeErrors") && BSON_ITER_HOLDS_ARRAY(&iter) && bson_iter_recurse(&iter, &child)) {
+		while (bson_iter_next(&child)) {
 			bson_t cbson;
 			uint32_t len;
 			const uint8_t *data;
@@ -305,11 +299,11 @@ PHP_METHOD(WriteResult, getWriteErrors)
 			zval *writeerror = NULL;
 #endif
 
-			if (!BSON_ITER_HOLDS_DOCUMENT(&iter)) {
+			if (!BSON_ITER_HOLDS_DOCUMENT(&child)) {
 				continue;
 			}
 
-			bson_iter_document(&iter, &len, &data);
+			bson_iter_document(&child, &len, &data);
 
 			if (!bson_init_static(&cbson, data, len)) {
 				continue;
@@ -423,12 +417,16 @@ static void php_phongo_writeresult_free_object(phongo_free_object_arg *object TS
 
 	zend_object_std_dtor(&intern->std TSRMLS_CC);
 
-	_mongoc_write_result_destroy(&intern->write_result);
+	if (intern->reply) {
+		bson_destroy(intern->reply);
+	}
 
 	if (intern->write_concern) {
 		mongoc_write_concern_destroy(intern->write_concern);
 	}
 
+	zval_ptr_dtor(&intern->manager);
+
 #if PHP_VERSION_ID < 70000
 	efree(intern);
 #endif
@@ -466,50 +464,95 @@ HashTable *php_phongo_writeresult_get_debug_info(zval *object, int *is_temp TSRM
 #else
 	zval                      retval = zval_used_for_init;
 #endif
-	php_phongo_bson_state     state = PHONGO_BSON_STATE_INITIALIZER;
-	bson_iter_t iter;
+	bson_iter_t iter, child;
 
 	intern = Z_WRITERESULT_OBJ_P(object);
 	*is_temp = 1;
 	array_init_size(&retval, 9);
 
-	ADD_ASSOC_LONG_EX(&retval, "nInserted", intern->write_result.nInserted);
-	ADD_ASSOC_LONG_EX(&retval, "nMatched", intern->write_result.nMatched);
-	if (intern->write_result.omit_nModified) {
-		ADD_ASSOC_NULL_EX(&retval, "nModified");
-	} else {
-		ADD_ASSOC_LONG_EX(&retval, "nModified", intern->write_result.nModified);
+#define SCP(field) \
+	if (bson_iter_init_find(&iter, intern->reply, (field)) && BSON_ITER_HOLDS_INT32(&iter)) { \
+		ADD_ASSOC_LONG_EX(&retval, (field), bson_iter_int32(&iter)); \
+	} else { \
+		ADD_ASSOC_NULL_EX(&retval, (field)); \
 	}
-	ADD_ASSOC_LONG_EX(&retval, "nRemoved", intern->write_result.nRemoved);
-	ADD_ASSOC_LONG_EX(&retval, "nUpserted", intern->write_result.nUpserted);
 
-	/* Use native arrays for debugging output */
-	state.map.root_type = PHONGO_TYPEMAP_NATIVE_ARRAY;
-	state.map.document_type = PHONGO_TYPEMAP_NATIVE_ARRAY;
+	SCP("nInserted");
+	SCP("nMatched");
+	SCP("nModified");
+	SCP("nRemoved");
+	SCP("nUpserted");
+#undef SCP
 
-	phongo_bson_to_zval_ex(bson_get_data(&intern->write_result.upserted), intern->write_result.upserted.len, &state);
-#if PHP_VERSION_ID >= 70000
-	ADD_ASSOC_ZVAL_EX(&retval, "upsertedIds", &state.zchild);
-#else
-	ADD_ASSOC_ZVAL_EX(&retval, "upsertedIds", state.zchild);
-#endif
-
-	phongo_bson_to_zval_ex(bson_get_data(&intern->write_result.writeErrors), intern->write_result.writeErrors.len, &state);
-#if PHP_VERSION_ID >= 70000
-	ADD_ASSOC_ZVAL_EX(&retval, "writeErrors", &state.zchild);
-#else
-	ADD_ASSOC_ZVAL_EX(&retval, "writeErrors", state.zchild);
-#endif
-
-	if (!bson_empty0(&intern->write_result.writeConcernErrors) &&
-			bson_iter_init(&iter, &intern->write_result.writeConcernErrors) &&
-			bson_iter_next(&iter) &&
-			BSON_ITER_HOLDS_DOCUMENT(&iter)) {
+	if (bson_iter_init_find(&iter, intern->reply, "upserted") && BSON_ITER_HOLDS_ARRAY(&iter)) {
 		uint32_t len;
 		const uint8_t *data;
+		php_phongo_bson_state state = PHONGO_BSON_STATE_INITIALIZER;
 
-		bson_iter_document(&iter, &len, &data);
+		/* Use native arrays for debugging output */
+		state.map.root_type = PHONGO_TYPEMAP_NATIVE_ARRAY;
+		state.map.document_type = PHONGO_TYPEMAP_NATIVE_ARRAY;
 
+		bson_iter_array(&iter, &len, &data);
+		phongo_bson_to_zval_ex(data, len, &state);
+#if PHP_VERSION_ID >= 70000
+		ADD_ASSOC_ZVAL_EX(&retval, "upsertedIds", &state.zchild);
+#else
+		ADD_ASSOC_ZVAL_EX(&retval, "upsertedIds", state.zchild);
+#endif
+	} else {
+#if PHP_VERSION_ID >= 70000
+		zval upsertedIds;
+		array_init(&upsertedIds);
+		ADD_ASSOC_ZVAL_EX(&retval, "upsertedIds", &upsertedIds);
+#else
+		zval *upsertedIds = NULL;
+		MAKE_STD_ZVAL(upsertedIds);
+		array_init(upsertedIds);
+		ADD_ASSOC_ZVAL_EX(&retval, "upsertedIds", upsertedIds);
+#endif
+	}
+
+	if (bson_iter_init_find(&iter, intern->reply, "writeErrors") && BSON_ITER_HOLDS_ARRAY(&iter)) {
+		uint32_t len;
+		const uint8_t *data;
+		php_phongo_bson_state state = PHONGO_BSON_STATE_INITIALIZER;
+
+		/* Use native arrays for debugging output */
+		state.map.root_type = PHONGO_TYPEMAP_NATIVE_ARRAY;
+		state.map.document_type = PHONGO_TYPEMAP_NATIVE_ARRAY;
+
+		bson_iter_array(&iter, &len, &data);
+		phongo_bson_to_zval_ex(data, len, &state);
+#if PHP_VERSION_ID >= 70000
+		ADD_ASSOC_ZVAL_EX(&retval, "writeErrors", &state.zchild);
+#else
+		ADD_ASSOC_ZVAL_EX(&retval, "writeErrors", state.zchild);
+#endif
+	} else {
+#if PHP_VERSION_ID >= 70000
+		zval writeErrors;
+		array_init(&writeErrors);
+		ADD_ASSOC_ZVAL_EX(&retval, "writeErrors", &writeErrors);
+#else
+		zval *writeErrors = NULL;
+		MAKE_STD_ZVAL(writeErrors);
+		array_init(writeErrors);
+		ADD_ASSOC_ZVAL_EX(&retval, "writeErrors", writeErrors);
+#endif
+	}
+
+	if (bson_iter_init_find(&iter, intern->reply, "writeConcernErrors") && BSON_ITER_HOLDS_ARRAY(&iter) &&
+	    bson_iter_recurse(&iter, &child) && bson_iter_next(&child) && BSON_ITER_HOLDS_DOCUMENT(&child)) {
+		uint32_t len;
+		const uint8_t *data;
+		php_phongo_bson_state state = PHONGO_BSON_STATE_INITIALIZER;
+
+		/* Use native arrays for debugging output */
+		state.map.root_type = PHONGO_TYPEMAP_NATIVE_ARRAY;
+		state.map.document_type = PHONGO_TYPEMAP_NATIVE_ARRAY;
+
+		bson_iter_document(&child, &len, &data);
 		phongo_bson_to_zval_ex(data, len, &state);
 #if PHP_VERSION_ID >= 70000
 		ADD_ASSOC_ZVAL_EX(&retval, "writeConcernError", &state.zchild);
@@ -525,15 +568,12 @@ HashTable *php_phongo_writeresult_get_debug_info(zval *object, int *is_temp TSRM
 		zval write_concern;
 
 		php_phongo_write_concern_to_zval(&write_concern, intern->write_concern);
+		ADD_ASSOC_ZVAL_EX(&retval, "writeConcern", &write_concern);
 #else
 		zval *write_concern = NULL;
 
 		MAKE_STD_ZVAL(write_concern);
 		php_phongo_write_concern_to_zval(write_concern, intern->write_concern);
-#endif
-#if PHP_VERSION_ID >= 70000
-		ADD_ASSOC_ZVAL_EX(&retval, "writeConcern", &write_concern);
-#else
 		ADD_ASSOC_ZVAL_EX(&retval, "writeConcern", write_concern);
 #endif
 	} else {
